#include <stdbool.h>
#include <stdio.h>

// needed for CreateSymbolicLink
// In case of problems, disable this
#define WINNTOVERWRITE
#ifdef WINNTOVERWRITE
	#undef _WIN32_WINNT
	#define _WIN32_WINNT 0x600
#endif
#include <windows.h>

#include "refs_prpc_callee_win.h"

////////////////////////////////////////////////////////////////////////////////

// NOTES:
// Handles could be backed up and used later, but by now closed immediately.
// So file open is not a sensible choice right now.

// state
static struct {
	DWORD offset;
	DWORD size;
	HANDLE handle;
	char name[MAX_PATH+1];
} fileLock;

//dummy state to avoid code optimization
static struct {
	CHAR* filename;
	DWORD attributes;
	DWORD filesize_low;
	DWORD filesize_high;
} listDirStruct;

////////////////////////////////////////////////////////////////////////////////

//
void winerrmsg(const char* msg) {
	LPVOID lpMsgBuf=NULL;
	LPVOID lpDisplayBuf=NULL;
	DWORD dw = GetLastError();
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER
		| FORMAT_MESSAGE_FROM_SYSTEM
		| FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dw,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,
		0, NULL );
	//
	errmsg2(msg, lpMsgBuf);
	//
	LocalFree(lpMsgBuf);
	LocalFree(lpDisplayBuf);
}

void snwinerrmsg(const char* msg, char* buff, size_t size) {
	LPVOID lpMsgBuf=NULL;
	LPVOID lpDisplayBuf=NULL;
	DWORD dw = GetLastError();
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER
		| FORMAT_MESSAGE_FROM_SYSTEM
		| FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dw,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf,
		0, NULL );
	//
	snerrmsg2(msg, lpMsgBuf, buff, size);
	//
	LocalFree(lpMsgBuf);
	LocalFree(lpDisplayBuf);
}

// just an alias for brevity
static HANDLE getHandle(char* filename)
{
	return CreateFileA(
		filename,
		GENERIC_READ|GENERIC_WRITE,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING, // set to CREATE_ALWAYS for ability to overwrite existing files
		FILE_ATTRIBUTE_NORMAL,
		NULL
		);
}

////////////////////////////////////////////////////////////////////////////////



int createDirectory(char* dirpath, SECURITY_ATTRIBUTES* securityAttributes)
{
	BOOL retval = CreateDirectoryA(dirpath, securityAttributes);
	if (!retval) {
		winerrmsg("createDirectory");
		return REFSCALL_GENERIC_ERROR;
	}
	return REFSCALL_SUCCESS;
}
//
int createFile(char* filename, uint32_t dwFlagsAndAttributes, SECURITY_ATTRIBUTES* securityAttributes)
{
	HANDLE handle = CreateFileA(
		filename,
		GENERIC_READ|GENERIC_WRITE,
		FILE_SHARE_DELETE|FILE_SHARE_READ|FILE_SHARE_WRITE,
		securityAttributes,
		CREATE_NEW, // set to CREATE_ALWAYS for ability to overwrite existing files
		dwFlagsAndAttributes,
		NULL
		);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("createFile");
		return REFSCALL_GENERIC_ERROR;
	}
	if (!CloseHandle(handle)) {
		winerrmsg("createFile-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int createSymLink(char* linkname, char* filename, uint32_t isdir)
{
#ifdef WINNTOVERWRITE
	uint32_t dwFlags = isdir ? 0x1 : 0x0;
	if (CreateSymbolicLinkA(linkname, filename, dwFlags))
		return REFSCALL_SUCCESS;
#endif
	winerrmsg("createSymLink");
	return REFSCALL_GENERIC_ERROR;
}
//
int copyFile(char* src, char* dst)
{
	if (CopyFile(src, dst, true))
		return REFSCALL_SUCCESS;
	winerrmsg("copyFile");
	return REFSCALL_GENERIC_ERROR;
}
//
int deleteFile(char* filename)
{
	if (DeleteFile(filename))
		return REFSCALL_SUCCESS;
	winerrmsg("deleteFile");
	return REFSCALL_GENERIC_ERROR;
}
//
int getCompressedFileSize(char* filename)
{
	uint32_t retval = GetCompressedFileSizeA(filename, NULL);
	if (retval != INVALID_FILE_SIZE)
		return REFSCALL_SUCCESS;
	winerrmsg("getCompressedFileSize");
	return REFSCALL_GENERIC_ERROR;
}
//
int getFileAttributes(char* filename)
{
	uint32_t retval = GetFileAttributesA(filename);
	if (retval != INVALID_FILE_ATTRIBUTES)
		return REFSCALL_SUCCESS;
	winerrmsg("getFileAttributes");
	return REFSCALL_GENERIC_ERROR;
}
//
int getFileSize(char* filename)
{
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("getFileSize-getHandle");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	uint32_t retval = GetFileSize(handle, NULL);
	//
	if (!CloseHandle(handle)) {
		winerrmsg("getFileSize-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	if (retval != INVALID_FILE_SIZE)
		return REFSCALL_SUCCESS;
	winerrmsg("getFileSize");
	return REFSCALL_GENERIC_ERROR;
}
//
int getFileTime(char* filename)
{
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE)
		return REFSCALL_HANDLE_OPEN_FAIL;
	// result not exposed by now
	FILETIME tcreat, taccess, twrite;
	if (!GetFileTime(handle, &tcreat, &taccess, &twrite)) {
		winerrmsg("getFileTime");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	//
	if (!CloseHandle(handle)) {
		winerrmsg("getFileTime-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int getFileType(char* filename)
{
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("getFileType-getHandle");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	// result not exposed by now
	uint32_t retval = GetFileType(handle);
	if (retval == FILE_TYPE_UNKNOWN && GetLastError() != NO_ERROR) {
		winerrmsg("getFileType");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	//
	if (!CloseHandle(handle)) {
		winerrmsg("getFileType-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int moveFile(char* src, char* dst)
{
	if (MoveFile(src, dst))
		return REFSCALL_SUCCESS;
	winerrmsg("moveFile");
	return REFSCALL_GENERIC_ERROR;
}
//
int lockFile(char* filename, uint32_t offset, uint32_t numbytes)
{
	if (fileLock.handle) {
		const char* premsg = "lockFile error for";
		const char* premsg2 = "already locked other file:";
		char buf[strlen(premsg)+MAX_PATH+strlen(premsg)+MAX_PATH+8];
		snprintf(buf, sizeof(buf), "%s %s. %s %s", premsg, filename, premsg2, fileLock.name);
		errmsg(buf);
		return REFSCALL_GENERIC_ERROR;
	}
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("lockFile-getHandle");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	BOOL retval = LockFile(
		handle,
		offset,
		0,
		numbytes,
		0);
	if (!retval) {
		winerrmsg("lockFile");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	// keep in state for unlocking
	strncpy(fileLock.name, filename, sizeof(fileLock.name)-1);
	fileLock.handle=handle;
	fileLock.offset=offset;
	fileLock.size=numbytes;
	//
	return REFSCALL_SUCCESS;
}
//
int removeDirectory(char* filename)
{
	if (RemoveDirectoryA(filename))
		return REFSCALL_SUCCESS;
	winerrmsg("removeDirectory");
	return REFSCALL_GENERIC_ERROR;
}
//
int setEndOfFile(char* filename, uint32_t offset)
{
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("setEndOfFile-getHandle");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	if (SetFilePointer(handle, offset, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER) {
		winerrmsg("setEndOfFile-SetFilePointer");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	if (!SetEndOfFile(handle)) {
		winerrmsg("setEndOfFile");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	if (!CloseHandle(handle)) {
		winerrmsg("setEndOfFile-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int setFileAttributes(char* filename, uint32_t attributes)
{
	if (SetFileAttributesA(filename, attributes))
		return REFSCALL_SUCCESS;
	winerrmsg("setFileAttributes");
	return REFSCALL_GENERIC_ERROR;
}
//
int setFileTime(char* filename, void* creat, void* access, void* write)
{
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("setFileTime-getHandle");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	if (!SetFileTime(handle, creat, access, write)) {
		winerrmsg("setFileTime");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	if (!CloseHandle(handle)) {
		winerrmsg("setFileTime-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int setFileValidData(char* filename, int64_t length)
{
	HANDLE handle = getHandle(filename);
	if (handle == INVALID_HANDLE_VALUE) {
		winerrmsg("setFileValidData-getHandle");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	if (!SetFileValidData(handle,length)) {
		winerrmsg("setFileValidData");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	if (!CloseHandle(handle)) {
		winerrmsg("setFileValidData-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int unlockFile(char* filename, uint32_t offset, uint32_t numbytes)
{
	// check for any file being locked
	if (!fileLock.handle) {
		const char* premsg = "unlockFile error for";
		const char* premsg2 = "No file locked.";
		char buf[strlen(premsg)+MAX_PATH+strlen(premsg)+MAX_PATH+8];
		snprintf(buf, sizeof(buf), "%s %s. %s", premsg, filename, premsg2);
		errmsg(buf);
		return REFSCALL_GENERIC_ERROR;
	}
	// check for unlock of locked file
	if (strcmp(fileLock.name, filename)
		|| fileLock.offset != offset
		|| fileLock.size != numbytes) {
		errmsg("unlockFile: data does not match locked file");
		return REFSCALL_GENERIC_ERROR;
	}
	// unlock
	HANDLE handle = fileLock.handle;
	BOOL retval = UnlockFile(
		handle,
		offset,
		0,
		numbytes,
		0);
	if (!retval) {
		winerrmsg("unlockFile");
		CloseHandle(handle);
		return REFSCALL_GENERIC_ERROR;
	}
	// cleanup
	fileLock.handle = NULL;
	if (!CloseHandle(handle)) {
		winerrmsg("unlockFile-CloseHandle");
		return REFSCALL_HANDLE_CLOSE_FAIL;
	}
	return REFSCALL_SUCCESS;
}
//
int readFile(char* filename)
{
	HANDLE hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		winerrmsg("readFile-CreateFile");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}

	char buffer[1024];
	DWORD bytesRead;
	// NOTE: by now no use of lpOverlapped
	if( FALSE == ReadFile(hFile, buffer, sizeof(buffer), &bytesRead, NULL) )
	{
		winerrmsg("readFile-ReadFile");
		CloseHandle(hFile);
		return REFSCALL_GENERIC_ERROR;
	}
	if (!CloseHandle(hFile)) {
		winerrmsg("readFile-CloseHandle");
		return REFSCALL_GENERIC_ERROR;
	}
	return REFSCALL_SUCCESS;
}

//
int listDirectory(char* dirpath)
{
	char szDir[MAX_PATH];
	WIN32_FIND_DATA ffd;
	HANDLE hFind = INVALID_HANDLE_VALUE;
	DWORD dwError=0;

	snprintf(szDir, sizeof(szDir), "%s\\*", dirpath);

	hFind = FindFirstFile(szDir, &ffd);

	if (INVALID_HANDLE_VALUE == hFind)
	{
		winerrmsg("listDirectory-FindFirstFile");
		return REFSCALL_HANDLE_OPEN_FAIL;
	}
	do
	{
		listDirStruct.attributes = ffd.dwFileAttributes;
		listDirStruct.filesize_low = ffd.nFileSizeLow;
		listDirStruct.filesize_high = ffd.nFileSizeHigh;
		listDirStruct.filename = ffd.cFileName;
	}
	while (FindNextFile(hFind, &ffd) != 0);

	dwError = GetLastError();
	if (dwError != ERROR_NO_MORE_FILES)
	{
		winerrmsg("listDirectory-FindNextFile");
		FindClose(hFind);
		return REFSCALL_GENERIC_ERROR;
	}

	FindClose(hFind);
	return REFSCALL_SUCCESS;
}

int createProcess(char* cmd)
{
	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	si.cb = sizeof(si);
	PROCESS_INFORMATION pi;
	ZeroMemory( &pi, sizeof(pi) );

	BOOL cpret = CreateProcess(
		NULL,
		cmd,
		NULL,
		NULL,
		FALSE,
		0,
		NULL,
		NULL,
		&si,
		&pi
		);
	if (!cpret) {
		char fmt[] = "CreateProcess Error for '%s'";
		char buf[strlen(fmt)+strlen(cmd)];
		snprintf(buf, sizeof(buf), fmt, cmd);
		errmsg(buf);
		return 1;
	}

	WaitForSingleObject(pi.hProcess, INFINITE);
	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
	return 0;
}

int copyDirectory(char* src, char* dst)
{
	char formatstring[] = "xcopy %s %s\\* /b /c /r /s /y";
	char cmd[strlen(formatstring)+strlen(src)+strlen(dst)+1];
	snprintf(cmd, sizeof(cmd), formatstring, src, dst);
	int ret = createProcess(cmd);
	if (ret)
		return REFSCALL_GENERIC_ERROR;
	return REFSCALL_SUCCESS;
}

int createJunction(char* junctionPath, char* targetPath)
{
	char formatstring[] = "cmd.exe /C mklink /J %s %s";
	char cmd[strlen(formatstring)+strlen(junctionPath)+strlen(targetPath)+1];
	snprintf(cmd, sizeof(cmd), formatstring, junctionPath, targetPath);
	int ret = createProcess(cmd);
	if (ret) {
		errmsg("createProcess");
		return REFSCALL_GENERIC_ERROR;
	}
	return REFSCALL_SUCCESS;
}

int copyJunction(char* srcJunction, char* dstJunction)
{
	/* TODO
	 * Step1: find taget of srcJunction
	 *	Net doable in a sane manner.
	 * Step2: createJunction(dstJunction, target)
	 */
	/*
	WIN32_FIND_DATAA fdata;
	HANDLE fffres = FindFirstFileA(srcJunction, &fdata);
	if (fffres == INVALID_HANDLE_VALUE) {
		winerrmsg("copyJunction-FindFirstFileA");
		return REFSCALL_GENERIC_ERROR;
	}
	// close handle
	if (!FindClose(fffres)) {
		winerrmsg("copyJunction-FindClose");
		return REFSCALL_GENERIC_ERROR;
	}
	// check for being a junction
	if (!(fdata.dwFileAttributes&FILE_ATTRIBUTE_REPARSE_POINT) || !(fdata.dwReserved0&IO_REPARSE_TAG_MOUNT_POINT)) {
		winerrmsg("copyJunction: srcJunction is no junction");
		return REFSCALL_GENERIC_ERROR;
	}
	*/
	errmsg("copyJunction NOT IMPLEMENTED YET!");
	return REFSCALL_GENERIC_ERROR;
}

