"""
Copyright (C) 2017 Sergej Schumilo

This file is part of kAFL Fuzzer (kAFL).

QEMU-PT is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

QEMU-PT is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with QEMU-PT.  If not, see <http://www.gnu.org/licenses/>.
"""

import mmap
import os
import sys
import resource
import select
import socket
import subprocess
import time
from socket import error as socket_error
import psutil
import mmh3
import shutil

from common.debug import log_qemu
from common.util import atomic_write, to_string_32_little, from_string_32_little, block_with_file

from common.util import Singleton
from multiprocessing import Process, Manager

import traceback


class QemuLookupSet:
    __metaclass__ = Singleton

    def __init__(self):
        manager = Manager()
        self.non_finding = manager.dict()
        self.crash = manager.dict()
        self.timeout = manager.dict()
        self.kasan = manager.dict()

class QemuMonitorException(Exception):
    pass

class WritePayloadException(Exception):
    pass

class qemu:
    SC_CLK_TCK = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

    def __init__(self, qid, config):
        self.qemu_out = None

        self.global_bitmap = None

        self.lookup = QemuLookupSet()

        self.bitmap_size = config.config_values['BITMAP_SHM_SIZE']
        self.config = config
        self.qemu_id = str(qid)

        self.process = None
        self.intervm_tty_write = None
        self.control = None
        self.control_fileno = None
        self.qemu_monitor = None

        self.first_image_set = True
        self.payload_filename   = "/dev/shm/kafl_qemu_payload_" + self.qemu_id
        self.payload2_filename   = "/dev/shm/kafl_qemu_payload2_" + self.qemu_id
        self.overlay_image_filename = "/dev/shm/kafl_qemu_overlay_image_" + self.qemu_id
        self.binary_filename    = "/dev/shm/kafl_qemu_binary_"  + self.qemu_id
        self.argv_filename      = "/dev/shm/kafl_argv_"         + self.qemu_id
        self.bitmap_filename    = "/dev/shm/kafl_bitmap_"       + self.qemu_id
        if self.config.argument_values.has_key('work_dir'):
            self.control_filename   = self.config.argument_values['work_dir'] + "/kafl_qemu_control_"  + self.qemu_id
            self.qemu_monitor_filename       = self.config.argument_values['work_dir'] + "/kafl_qemu_monitor_" + self.qemu_id
        else:
            self.control_filename   = "/tmp/kafl_qemu_control_"  + self.qemu_id
            self.qemu_monitor_filename       = "/tmp/kafl_qemu_monitor_" + self.qemu_id
        self.start_ticks = 0
        self.end_ticks = 0
        self.tick_timeout_treshold = self.config.config_values["DEFAULT_TICK_TIMEOUT"]

        self.cmd =  self.config.config_values['QEMU_KAFL_LOCATION'] + " " \
                    "-hdb " + self.config.argument_values['ram_file'] + " " \
                    "-hda " + self.config.argument_values['overlay_dir'] +  "/overlay_" + self.qemu_id + ".qcow2 " \
                    "-serial mon:stdio " \
                    "-enable-kvm " \
                    "-k de " \
                    "-m " + str(config.argument_values['mem']) + " " \
                    "-net user " \
                    "-net nic " \
                    "-monitor unix:" + self.qemu_monitor_filename + ",server,nowait "\
                    "-chardev socket,server,nowait,path=" + self.control_filename + \
                    ",id=kafl_interface " \
                    "-device kafl,chardev=kafl_interface,bitmap_size=" + str(self.bitmap_size) + ",shm0=" + self.binary_filename + \
                    ",shm1=" + self.payload_filename + \
                    ",shm2=" + self.payload2_filename + \
                    ",bitmap=" + self.bitmap_filename


        for i in range(1):
            key = "ip" + str(i)
            if self.config.argument_values.has_key(key) and self.config.argument_values[key]:
                range_a = hex(self.config.argument_values[key][0]).replace("L", "")
                range_b = hex(self.config.argument_values[key][1]).replace("L", "") 
                self.cmd += ",ip" + str(i) + "_a=" + range_a + ",ip" + str(i) + "_b=" + range_b
                self.cmd += ",filter" + str(i) + "=/dev/shm/kafl_filter" + str(i)

        graphics = False
        if graphics:
            self.cmd += " -vnc :1,password"
        else:
            self.cmd += " -nographic"

        self.cmd += " -loadvm " + self.config.argument_values["S"] + " "

        if self.config.argument_values["macOS"]:
            self.cmd = self.cmd.replace("-net user -net nic", "-netdev user,id=hub0port0 -device e1000-82545em,netdev=hub0port0,id=mac_vnet0 -cpu Penryn,kvm=off,vendor=GenuineIntel -device isa-applesmc,osk=\"" + self.config.config_values["APPLE-SMC-OSK"].replace("\"", "") + "\" -machine pc-q35-2.4")
            if qid == 0:
                self.cmd = self.cmd.replace("-machine pc-q35-2.4", "-machine pc-q35-2.4 -redir tcp:5901:0.0.0.0:5900 -redir tcp:10022:0.0.0.0:22")
        else:
            self.cmd += " -machine pc-i440fx-2.6 "

        self.kafl_shm_f = None
        self.kafl_shm   = None
        self.fs_shm_f   = None
        self.fs_shm     = None
        self.fs_shm2_f   = None
        self.fs_shm2    = None

        self.payload_shm_f   = None
        self.payload_shm     = None

        self.bitmap_shm_f   = None
        self.bitmap_shm     = None

        self.e = select.epoll()
        self.crashed = False
        self.timeout = False
        self.kasan = False
        self.shm_problem = False
        self.initial_mem_usage = 0

        self.stat_fd = None

        if qid == 0:
            log_qemu("Launching Virtual Maschine...CMD:\n" + self.cmd, self.qemu_id)
        else:
            log_qemu("Launching Virtual Maschine...", self.qemu_id)
        self.virgin_bitmap = ''.join(chr(0xff) for x in range(self.bitmap_size))

        self.__set_binary(self.binary_filename, self.config.argument_values['executable'], (16 << 20))


    def __del__(self):
        if self.process:
            log_qemu("DEL QEMU pid {}".format(self.process.pid), self.qemu_id)
            os.system("kill -9 " + str(self.process.pid))

        try:
            if self.process:
                try:
                    self.process.kill()
                    self.process = None
                except Exception as e:
                    log_qemu("Error in killing qemu {}".format(str(e)), self.qemu_id)

            if self.e:
                if self.control_fileno:
                    self.e.unregister(self.control_fileno)
                    self.e = None

            if self.intervm_tty_write:
                self.intervm_tty_write.close()
                self.intervm_tty_write = None
            if self.control:
                self.control.close()
                self.control = None
            if self.qemu_monitor:
                self.qemu_monitor.close()
                self.qemu_monitor = None
        except OSError:
            pass

        try:
            self.kafl_shm.close()
            self.kafl_shm = None
        except:
            pass

        try:
            self.qemu_out.close()
            self.qemu_out = None
        except:
            pass

        try:
            self.fs_shm.close()
            self.fs_shm = None
        except:
            pass

        try:
            self.fs_shm2.close()
            self.fs_shm2 = None
        except:
            pass

        try:
            if self.kafl_shm_f:
                os.close(self.kafl_shm_f)
                self.kafl_shm_f = None
        except:
            pass

        try:
            if self.fs_shm_f:
                os.close(self.fs_shm_f)
                self.fs_shm_f = None
        except:
            pass

        try:
            if self.fs_shm2_f:
                os.close(self.fs_shm2_f)
                self.fs_shm2_f = None
        except:
            pass

        try:
            if self.stat_fd:
                self.stat_fd.close()
                self.stat_fd = None
        except:
            pass

        try:
            if self.global_bitmap:
                self.global_bitmap.close()
                self.global_bitmap = None
        except:
            pass

        try:
            if self.global_bitmap_fd:
                os.close(self.global_bitmap_fd)
                self.global_bitmap_fd = None
        except:
            pass


    def __get_pid_guest_ticks(self):
        if self.stat_fd:
            self.stat_fd.seek(0)
            self.stat_fd.flush()
            return int(self.stat_fd.readline().split(" ")[42])
        return 0

    def __set_binary(self, filename, binaryfile, max_size):
        shm_fd = os.open(filename, os.O_RDWR | os.O_SYNC | os.O_CREAT)
        os.ftruncate(shm_fd, max_size)
        shm = mmap.mmap(shm_fd, max_size, mmap.MAP_SHARED, mmap.PROT_WRITE | mmap.PROT_READ)
        shm.seek(0x0)
        shm.write('\x00' * max_size)
        shm.seek(0x0)

        f = open(binaryfile, "rb")
        bytes = f.read(1024)
        if bytes:
            shm.write(bytes)
        while bytes != "":
            bytes = f.read(1024)
            if bytes:
                shm.write(bytes)

        f.close()
        shm.close()
        os.close(shm_fd)

    def __set_vnc_password(self):
        ack = subprocess.check_output(["./qm_wrapper.sh", self.qemu_monitor_filename, "change vnc password\n200190"])


    def __delete_usb_drive(self):
        #for debugging we only wait for file exist
        # import time
        # log_qemu("please delete usb drive", self.qemu_id)
        # while True:
        #     if os.path.exists("/tmp/dbg_lock"):
        #         os.remove("/tmp/dbg_lock")
        #         return
        #     time.sleep(1)

        ack = subprocess.check_output(["./qm_wrapper.sh", self.qemu_monitor_filename, "device_del fuzz_device"])
        ack = ack.split('\x1b[K')
        ack = ack[-1].strip()
        log_qemu("device_del response: {}".format(ack), self.qemu_id)
        if "(qemu)" not in ack:
            raise QemuMonitorException("device_del command not successful")


    def __add_usb_drive(self):
        #for debugging we only wait for file exist
        # import time
        # log_qemu("please add usb drive", self.qemu_id)
        # while True:
        #     if os.path.exists("/tmp/dbg_lock"):
        #         os.remove("/tmp/dbg_lock")
        #         return
        #     time.sleep(1)

        # cache=none does not work if qcow file is placed in shm: "file system may not support O_DIRECT"
        # drive_add_cmd = "drive_add 5 if=none,file={},format=qcow2,id=fuzz_disk,cache=none\n".format(self.overlay_image_filename)
        drive_add_cmd = "drive_add 5 if=none,file={},format=qcow2,id=fuzz_disk,cache=writethrough".format(self.overlay_image_filename)
        #log_qemu("send {}".format(drive_add_cmd), self.qemu_id)
        #
        ack = subprocess.check_output(["./qm_wrapper.sh", self.qemu_monitor_filename, drive_add_cmd])
        ack = ack.split('\x1b[K')
        ack = ack[-1].strip()
        log_qemu("drive add response: {}".format(ack), self.qemu_id)
        if "OK" not in ack:
            raise QemuMonitorException("drive_add command not successful")

        device_add_cmd = "device_add usb-storage,drive=fuzz_disk,id=fuzz_device"
        ack = subprocess.check_output(["./qm_wrapper.sh", self.qemu_monitor_filename, device_add_cmd])
        ack = ack.split('\x1b[K')
        ack = ack[-1].strip()
        log_qemu("device add response: {}".format(ack), self.qemu_id)
        if "(qemu)" not in ack:
            raise QemuMonitorException("device_add command not successful")


    def set_tick_timeout_treshold(self, treshold):
        self.tick_timeout_treshold = treshold


    def start(self, verbose=False):
        if verbose:
            self.process = subprocess.Popen(filter(None, self.cmd.split(" ")),
                                            stdin=None,
                                            stdout=None,
                                            stderr=None)
        else:
            self.qemu_out = open("qemu{}.log".format(self.qemu_id), "a+")
            self.process = subprocess.Popen(filter(None, self.cmd.split(" ")),
                                            stdin=subprocess.PIPE,
                                            stdout=self.qemu_out,
                                            stderr=self.qemu_out)


        # block_with_file("VNC")
        self.stat_fd = open("/proc/" + str(self.process.pid) + "/stat")
        self.init()
        try:
            self.set_init_state()
        except Exception as e:
            log_qemu("set init state failed {}".format(str(e)), self.qemu_id)
            return False
        self.initial_mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        #time.sleep(1)
        self.kafl_shm.seek(0x0)
        self.kafl_shm.write(self.virgin_bitmap)
        self.kafl_shm.flush()
        return True

    def set_init_state(self):
        # since we've reloaded the VM - let's assume there is no panic state...
        self.crashed = False
        self.timeout = False
        self.kasan = False
        self.start_ticks = 0
        self.end_ticks = 0
        self.control.settimeout(10.0)
        v = self.control.recv(1)
        log_qemu("Initial stage 1 handshake ["+ str(v) + "] done...", self.qemu_id)
        self.__set_binary(self.binary_filename, self.config.argument_values['executable'], (16 << 20))
        if v != 'D':
            self.control.send('R')
            v = self.control.recv(1)
            log_qemu("Initial stage 2 handshake ["+ str(v) + "] done...", self.qemu_id)
        v = self.control.recv(1)
        # should be R for requesting fuzzing program
        if v != 'R':
            log_qemu("Initial handshake failure. received [{}]".format(str(v)), self.qemu_id)
            raise Exception("Init protocol fail")
        # program is allready written, we can release guest
        self.control.send('R')
        # guest should now request next payload
        v = self.control.recv(1)
        log_qemu("Initial stage 3 handshake ["+ str(v) + "] done...", self.qemu_id)
        self.control.settimeout(5.0)
        # self.__set_vnc_password()


    def init(self):
        self.control = socket.socket(socket.AF_UNIX)
        while True:
            try:
                self.control.connect(self.control_filename)
                break
            except socket_error:
                pass
                #time.sleep(0.01)

        self.kafl_shm_f     = os.open(self.bitmap_filename, os.O_RDWR | os.O_SYNC | os.O_CREAT)
        self.fs_shm_f       = os.open(self.payload_filename, os.O_RDWR | os.O_SYNC | os.O_CREAT)
        self.fs_shm2_f       = os.open(self.payload2_filename, os.O_RDWR | os.O_SYNC | os.O_CREAT)

        #argv_fd             = os.open(self.argv_filename, os.O_RDWR | os.O_SYNC | os.O_CREAT)
        os.ftruncate(self.kafl_shm_f, self.bitmap_size)
        os.ftruncate(self.fs_shm_f, (128 << 10))
        os.ftruncate(self.fs_shm2_f, (1 << 20))
        #os.ftruncate(argv_fd, (4 << 10))

        self.kafl_shm       = mmap.mmap(self.kafl_shm_f, self.bitmap_size, mmap.MAP_SHARED, mmap.PROT_WRITE | mmap.PROT_READ)
        self.fs_shm         = mmap.mmap(self.fs_shm_f, (128 << 10),  mmap.MAP_SHARED, mmap.PROT_WRITE | mmap.PROT_READ)
        self.fs_shm2        = mmap.mmap(self.fs_shm2_f, (1 << 20),  mmap.MAP_SHARED, mmap.PROT_WRITE | mmap.PROT_READ)


        return True

    def soft_reload(self):
        self.crashed = False
        self.timeout = False
        self.kasan = False
        self.start_ticks = 0
        self.end_ticks = 0
        self.control.settimeout(2.0)


        again = True
        while True:
            try:
                # send L for reload request
                self.control.send('L')
                while True:
                    # guest should return L for ack reload
                    if self.control.recv(1) == 'L':
                        break
                # here we should receive 'D' for first handshake release
                v = self.control.recv(1)
                break
            # sometimes L fails on first try so we get a timeout and have to try it again
            except Exception as e:
                tb = traceback.format_exc()
                log_qemu("soft reload exception {}, {}".format(str(e), tb), self.qemu_id)
                if not again:
                    raise e
                again = False

        self.control.settimeout(10.0)
        self.__set_binary(self.binary_filename, self.config.argument_values['executable'], (16 << 20))
        if v != 'D':
            self.control.send('R')
            v = self.control.recv(1)

        v = self.control.recv(1)
        # should be R for requesting fuzzing program
        if v != 'R':
            log_qemu("Initial handshake failure. received [{}]".format(str(v)), self.qemu_id)
            raise Exception("Init protocol fail")
        # program is allready written, we can release guest
        self.control.send('R')
        #v = self.control.recv(1)
        v = self.control.recv(1)
        self.control.settimeout(5.0)

    # Return Codes:
    # 0 : OK
    # 1 : Crash
    # 2 : Timeout
    # 3 : KASAN
    def check_recv(self, timeout_detection=True):
        if timeout_detection:
            self.control.settimeout(1.25)
        try:
            result = self.control.recv(1)
        except socket_error:
            return 2

        if result == 'C':
            return 1
        elif result == 'K':
            return 3
        elif result == 'R':
            return 0
            log_qemu("Finding...Type is ["+ result + "]", self.qemu_id)
        elif result == 'L':
            log_qemu("Failed to write payload", self.qemu_id)
            raise WritePayloadException("can not write payload shared mem")
        return 2

    def send_payload(self, timeout_detection=True):
        log_qemu("Send Payload", self.qemu_id)
        self.start_ticks = self.__get_pid_guest_ticks()
        try:
            self.control.send("R")
        except OSError:
            log_qemu("Failed to send payload... canot use control file", self.qemu_id)
            return None

        self.crashed = False
        self.timeout = False
        self.kasan = False

        if timeout_detection:
            counter = 0
            while True:
                value = self.check_recv()
                if value == 2:
                    self.end_ticks = self.__get_pid_guest_ticks()
                    if (self.end_ticks-self.start_ticks) >= self.tick_timeout_treshold:
                        log_qemu("Tick timeout {}".format(self.tick_timeout_treshold), self.qemu_id)
                        break
                    if counter >= 100:
                        log_qemu("Counter timeout", self.qemu_id)
                        break
                    counter += 1
                else:
                    break
            self.end_ticks = self.__get_pid_guest_ticks()
        else:
            value = self.check_recv(timeout_detection=False)

        if value == 1:
            self.crashed = True
            self.finalize_iteration()
        elif value == 2:
            self.timeout = True
            self.finalize_iteration()
        elif value == 3:
            self.kasan = True
            self.finalize_iteration()
        self.kafl_shm.seek(0x0)
        return self.kafl_shm.read(self.bitmap_size)

    def enable_sampling_mode(self):
        self.control.send("S")

    def disable_sampling_mode(self):
        self.control.send("O")

    def submit_sampling_run(self):
        self.control.send("T")

    # in master shm payload file, the size is allready encoded
    # here he have not need to write the size explicitaly
    # in set payload we have to write size expl.
    def copy_master_payload(self, shm, num, size):
        shm.seek(size * num)
        size1_bytes = shm.read(4)
        size1 = from_string_32_little(size1_bytes)
        size2_bytes = shm.read(4)
        size2 = from_string_32_little(size2_bytes)
        if size1 > 0:
            payload1 = shm.read(size1)
        else:
            payload1 = ""
        if size2 > 0:
            payload2 = shm.read(size2)
        else:
            payload2 = ""

        if size2 > 0:
            self.set_payload(payload2)

        if size1 > 0:
            self.set_image(payload1)

        return payload1, payload2

    def copy_mapserver_payload(self, shm, num, size, payload, payload2):
        shm.seek(size * num)
        psize1 = len(payload)
        psize2 = len(payload2)
        shm.write(to_string_32_little(psize1))
        shm.write(to_string_32_little(psize2))
        shm.write(payload)
        shm.write(payload2)

    def open_global_bitmap(self):
        self.global_bitmap_fd = os.open(self.config.argument_values['work_dir'] + "/bitmap", os.O_RDWR | os.O_SYNC | os.O_CREAT)
        os.ftruncate(self.global_bitmap_fd, self.bitmap_size)
        self.global_bitmap = mmap.mmap(self.global_bitmap_fd, self.bitmap_size, mmap.MAP_SHARED, mmap.PROT_WRITE | mmap.PROT_READ)

    def verifiy_input(self, payload, payload2, bitmap, runs=3):
        crashed = self.crashed
        timeout = self.timeout
        kasan = self.kasan
        failed = False
        try:
            self.enable_sampling_mode()
            init = True
            tmp_bitmap1 = bitmap
            for i in range(runs):
                self.set_image(payload)
                self.set_payload(payload2)

                if not init:
                    tmp_bitmap1 = self.send_payload(timeout_detection=False)
                    if (self.crashed or self.kasan or self.timeout):
                        break
                    else:
                        self.submit_sampling_run()

                tmp_bitmap2 = self.send_payload(timeout_detection=False)
                if (self.crashed or self.kasan or self.timeout):
                    break
                else:
                    self.submit_sampling_run()
                if tmp_bitmap1 == tmp_bitmap2:
                    break
                init = False
                
        except:
            failed = True

        self.crashed = crashed or self.crashed
        self.timeout = timeout or self.timeout
        self.kasan = kasan or self.kasan

        try:
            if not self.timeout:
                self.submit_sampling_run()
            self.disable_sampling_mode()
            if not failed:
                return tmp_bitmap2
            else:
                return bitmap            
        except:
            self.timeout = True
            return bitmap

    def check_for_unseen_bits(self, bitmap):
        if not self.global_bitmap:
            self.open_global_bitmap()

        for i in range(self.bitmap_size):
            if bitmap[i] != '\xff':
                if self.global_bitmap[i] == '\x00': # REVIEW why should this not only the case if bitmap is non 0 at this offset?
                    return True
                if (ord(bitmap[i]) | ord(self.global_bitmap[i])) != ord(self.global_bitmap[i]):
                    return True
        return False

    
    def copy_bitmap(self, shm, num, size, bitmap, payload, payload2, effector_mode=False):
        new_hash = mmh3.hash64(bitmap)
        if not (self.crashed or self.kasan or self.timeout):
            if new_hash in self.lookup.non_finding:
                if effector_mode:
                    shm.seek(size * num)
                    shm.write(bitmap)
                    return True
                else:
                    shm.seek((size * num) + len(bitmap))
                    return False

        if not (self.crashed or self.kasan or self.timeout) and not self.check_for_unseen_bits(bitmap):
            self.lookup.non_finding[new_hash] = None
            return False
        if not (self.timeout):
            bitmap = self.verifiy_input(payload, payload2, bitmap)
            shm.seek(size * num)
            shm.write(bitmap)
        self.lookup.non_finding[new_hash] = None
        return True

    def set_payload(self, payload):
        plength = to_string_32_little(len(payload))
        self.fs_shm.seek(0)
        self.fs_shm.write(plength)
        self.fs_shm.write(payload)

    def set_payload2(self, payload):
        plength = to_string_32_little(len(payload))
        self.fs_shm2.seek(0)
        self.fs_shm2.write(plength)
        self.fs_shm2.write(payload)

    def set_image(self, payload):
        self.set_payload2(payload)
        return

    def finalize_iteration(self):
        try:
            self.control.send('F')
            self.control.recv(1)
        except:
            log_qemu("finalize_iteration failed...", self.qemu_id)

